/*
 * Copyright 2017 ObjectBox Ltd. All rights reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.objectbox.android;

import android.Manifest;
import android.app.Notification;
import android.app.Notification.Builder;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.net.Uri;
import android.os.Build.VERSION;
import android.os.Build.VERSION_CODES;
import android.os.Process;
import android.util.Log;

import io.objectbox.BoxStore;

import static android.content.Intent.FLAG_ACTIVITY_NEW_TASK;

public class AndroidObjectBrowser {

    private static final String TAG = "ObjectBrowser";
    private static final String NOTIFICATION_CHANNEL_ID = "objectbox-browser";

    private final BoxStore boxStore;
    private int notificationId;

    public AndroidObjectBrowser(BoxStore boxStore) {
        this.boxStore = boxStore;
    }

    public int getNotificationId() {
        return notificationId;
    }

    public void setNotificationId(int notificationId) {
        this.notificationId = notificationId;
    }

    /**
     * Starts ObjectBox data browser and displays a notification to launch the browser view. When
     * tapped, launches a foreground service to keep the app alive. Stop it from its notification.
     */
    public boolean start(Context context) {
        if (!BoxStore.isObjectBrowserAvailable()) {
            return false;
        }

        // SecurityException if no INTERNET permission
        // better than adding it through the AAR manifest (added even if data browser is not used)
        context.enforcePermission(Manifest.permission.INTERNET, Process.myPid(), Process.myUid(), null);

        int alreadyRunningPort = boxStore.getObjectBrowserPort();
        if (alreadyRunningPort != 0) {
            Log.w(TAG, "ObjectBrowser is already running at port " + alreadyRunningPort);
            return false;
        }
        String url = boxStore.startObjectBrowser();
        if (url == null) {
            return false;
        }
        Log.i(TAG, "ObjectBrowser started: " + url);
        int port = boxStore.getObjectBrowserPort();
        Log.i(TAG, "Command to forward ObjectBrowser to connected host: adb forward tcp:" + port + " tcp:" + port);


        if (notificationId == 0) {
            notificationId = 19770000 + port;
        }

        // build intent and show notification
        Intent intent = new Intent(context, AndroidObjectBrowserReceiver.class);
        intent.setAction(AndroidObjectBrowserReceiver.ACTION_KEEP_ALIVE);
        intent.putExtra(AndroidObjectBrowserService.EXTRA_KEY_URL, url);
        intent.putExtra(AndroidObjectBrowserService.EXTRA_KEY_PORT, port);
        intent.putExtra(AndroidObjectBrowserService.EXTRA_KEY_NOTIFICATION_ID, notificationId);
        PendingIntent pendingIntent = PendingIntent.getBroadcast(context, 0, intent,
                PendingIntent.FLAG_CANCEL_CURRENT);

        NotificationManager manager =
                (NotificationManager) context.getSystemService(Context.NOTIFICATION_SERVICE);
        Builder builder = buildBaseNotification(context, port, manager);
        builder.setContentIntent(pendingIntent);
        if (manager != null) {
            manager.notify(notificationId, builder.getNotification());
        }

        return true;
    }

    static Builder buildBaseNotification(Context context, int port, NotificationManager manager) {
        if (VERSION.SDK_INT >= VERSION_CODES.O) {
            NotificationChannel channel = new NotificationChannel(NOTIFICATION_CHANNEL_ID,
                    "ObjectBox Browser", NotificationManager.IMPORTANCE_DEFAULT);
            // if channel already exists, create call will be ignored
            manager.createNotificationChannel(channel);
        }

        Notification.Builder builder;
        if (VERSION.SDK_INT >= VERSION_CODES.O) {
            builder = new Notification.Builder(context, NOTIFICATION_CHANNEL_ID);
        } else {
            builder = new Notification.Builder(context);
        }

        builder.setContentTitle(context.getString(R.string.objectbox_objectBrowserNotificationTitle))
                .setContentText(context.getString(R.string.objectbox_objectBrowserNotificationText, port))
                .setSmallIcon(R.drawable.objectbox_notification);

        return builder;
    }

    static Intent viewIntent(String url) {
        Intent viewIntent = new Intent(Intent.ACTION_VIEW, Uri.parse(url));
        viewIntent.setFlags(FLAG_ACTIVITY_NEW_TASK);
        return viewIntent;
    }
}
